import numpy as np
import numba

@numba.njit()
def project_xyz(x: np.float64, y: np.float64, t: np.float64, vx: np.float64, vy: np.float64, reference_time: np.float64):
    x_prime = x - vx * (t - reference_time)
    y_prime = y - vy * (t - reference_time)
    return x_prime, y_prime

@numba.njit()
def digitize_point(x, y, min_x, min_y, dx, dy):
    return int((x - min_x)/dx), int((y - min_y)/dy)



import numpy as np
import numba
from .primitives import project_xyz, digitize_point

@numba.njit(parallel=True)
def projected_bounds(X: np.ndarray, directions: np.ndarray, reference_time: np.float64):
    n, d = X.shape
    num_dir = directions.shape[0]
    min_x_arr = np.full(num_dir, np.inf)
    max_x_arr = np.full(num_dir, -np.inf)
    min_y_arr = np.full(num_dir, np.inf)
    max_y_arr = np.full(num_dir, -np.inf)
    
    for dir_idx in numba.prange(num_dir): # for each direction
        if dir_idx < num_dir:
            direction = directions[dir_idx]
            for i in range(n): # for each point
                p = X[i]
                x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
                min_x_arr[dir_idx] = min(min_x_arr[dir_idx], x_prime)
                max_x_arr[dir_idx] = max(max_x_arr[dir_idx], x_prime)
                min_y_arr[dir_idx] = min(min_y_arr[dir_idx], y_prime)
                max_y_arr[dir_idx] = max(max_y_arr[dir_idx], y_prime)

    min_x = np.min(min_x_arr)
    max_x = np.max(max_x_arr)
    min_y = np.min(min_y_arr)
    max_y = np.max(max_y_arr)
    
    return min_x, max_x, min_y, max_y


def hough_max(hough):
    idx = np.unravel_index(hough.argmax(), hough.shape)
    return idx, hough[idx]

@numba.njit(parallel=True)
def make_bins(X: np.ndarray, directions: np.ndarray, x_min, y_min, dx, dy, reference_time):
    n, d = X.shape
    num_dir = directions.shape[0]
    bins = np.full((n, num_dir, 2), -1)
    for dir_idx in numba.prange(num_dir): # for each direction
        if dir_idx < num_dir:
            direction = directions[dir_idx]
            for i in range(n): # for each point
                p = X[i]
                x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
                x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)
                bins[i, dir_idx, 0] = x_idx
                bins[i, dir_idx, 1] = y_idx
    return bins

@numba.njit(parallel=True)
def vote_points(hough: np.ndarray, X: np.ndarray, directions: np.ndarray, x_min, y_min, dx, dy, reference_time, coef):
    n, d = X.shape
    num_dir = directions.shape[0]
    for dir_idx in numba.prange(num_dir): # for each direction
        if dir_idx < num_dir:
            direction = directions[dir_idx]
            for i in range(n): # for each point
                p = X[i]
                x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
                x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)
                if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
                    hough[dir_idx, x_idx, y_idx] += coef
    return hough

@numba.njit(parallel=True)
def vote_bins(hough: np.ndarray, bins: np.ndarray, coef):
    n, num_dir, _ = bins.shape
    for dir_idx in numba.prange(num_dir): # for each direction
        if dir_idx < num_dir:
            for i in range(n): # for each point
                x_idx = bins[i, dir_idx, 0]
                y_idx = bins[i, dir_idx, 1]
                if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
                    hough[dir_idx, x_idx, y_idx] += coef
    return hough

@numba.njit(parallel=True)
def find_voters_bins(hough: np.ndarray, bins: np.ndarray, mask_dir, mask_x, mask_y):
    n, num_dir, _ = bins.shape
    
    num_threads = numba.get_num_threads()
    mask = np.full((num_threads, n), False)
    _mask = np.full(n, False)
    for dir_idx in numba.prange(num_dir): # for each direction
        thread_idx = dir_idx % num_threads
        if dir_idx < num_dir:
            for i in range(n): # for each point
                x_idx = bins[i, dir_idx, 0]
                y_idx = bins[i, dir_idx, 1]                
                if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
                    if dir_idx == mask_dir and x_idx == mask_x and y_idx == mask_y:
                        mask[thread_idx, i] |= True
    for i in range(n):
        for dir_idx in range(num_dir):
            thread_idx = dir_idx % num_threads
            _mask[i] |= mask[thread_idx, i]
            
    return _mask

@numba.njit(parallel=True)
def find_voters_points(hough: np.ndarray, X: np.ndarray, directions: np.ndarray, x_min, y_min, dx, dy, reference_time, mask_dir, mask_x, mask_y):
    n, d = X.shape
    num_dir = directions.shape[0]
    num_threads = numba.get_num_threads()
    
    mask = np.full((num_threads, n), False)
    _mask = np.full(n, False)
    for dir_idx in numba.prange(num_dir): # for each direction
        thread_idx = dir_idx % num_threads
        if dir_idx < num_dir:
            direction = directions[dir_idx]
            for i in range(n): # for each point
                p = X[i]
                x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
                x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)
                if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
                    if dir_idx == mask_dir and x_idx == mask_x and y_idx == mask_y:
                        mask[thread_idx, i] = True
    for i in range(n):
        for thread_idx in range(num_threads):
            _mask[i] |= mask[thread_idx, i]
            
    return _mask




from .directions import SearchDirections
from .postprocess import refine, gather

import numpy as np
from numba import cuda
import astropy.table

import astropy.units as u
from pathlib import Path
import logging

logging.basicConfig()
log = logging.getLogger(__name__)

def search_gpu(X, directions, dx, reference_time, num_results=10):
    from .gpu_impl import projected_bounds, _vote_points, _vote_points_mask, _find_voters_points, _hough_max
    n = X.shape[0]
    
    x_min, x_max, y_min, y_max = projected_bounds(X, directions.b, reference_time)
    
    num_dir = directions.b.shape[0]
    _dx = dx.to(u.deg).value
    _dy = _dx
    num_x = int((x_max - x_min) / _dx  + 1)
    num_y = int((y_max - y_min) / _dy  + 1)

    log.info("creating hough space with shape (%d, %d, %d)", num_dir, num_x, num_y)
    hough = np.zeros((num_dir, num_x, num_y), dtype=np.int32)

    num_dir, num_x, num_y = hough.shape
    
    mask = np.full((num_dir, n), False)
    
    d_hough = cuda.to_device(hough)
    d_mask = cuda.to_device(mask)
    d_X = cuda.to_device(X)
    d_directions = cuda.to_device(directions.b)
    d_max = cuda.to_device(np.zeros((hough.shape[0], 3), dtype=np.int32))
    d_results = cuda.to_device(np.zeros((num_results, 4), dtype=np.int32))

    def _vote(coef=1):
        # Configure GPU threads and blocks
        threads_per_block = (16, 16)  # Tunable parameters
        blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                           (n + threads_per_block[1] - 1) // threads_per_block[1])

        # Launch the CUDA kernel
        _vote_points[blocks_per_grid, threads_per_block](
            d_hough, d_X, d_directions, x_min, y_min, _dx, _dy, reference_time, coef
        )

    def _vote_mask(coef=1):
        # Configure GPU threads and blocks
        threads_per_block = (16, 16)  # Tunable parameters
        blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                           (n + threads_per_block[1] - 1) // threads_per_block[1])

        # Launch the CUDA kernel
        _vote_points_mask[blocks_per_grid, threads_per_block](
            d_hough, d_mask, d_X, d_directions, x_min, y_min, _dx, _dy, reference_time, coef
        )
    
    def _max():
        _hough_max[256, num_dir // 256 + 1](d_max, d_hough)
        
    def _find(mask_dir, mask_x, mask_y):
        threads_per_block = (16, 16)  # Tunable parameters
        blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                           (n + threads_per_block[1] - 1) // threads_per_block[1])

        _find_voters_points[blocks_per_grid, threads_per_block](
            d_X, d_directions, d_mask, x_min, y_min, _dx, _dy, reference_time, mask_dir, mask_x, mask_y
        )

    _vote(coef=1)
    # results = [] # cuda device array
    for n_i in range(num_results):
        _max()
        i = -1
        v = -np.inf
        for _ in range(len(d_max)):
            if d_max[_, 2] > v:
                v = d_max[_, 2]
                i = _
        j = d_max[i, 0]
        k = d_max[i, 1]
        d_results[n_i, 0] = i
        d_results[n_i, 1] = j
        d_results[n_i, 2] = k
        d_results[n_i, 3] = v
        _find(i, j, k)
        print("cluster has value", v, "at", (i, j, k))
        _vote_mask(coef=-1)
    
    return d_results.copy_to_host()

def search(X, directions, dx, reference_time, num_results=10, precompute=False, gpu=False):
    if gpu:
        from .gpu_impl import projected_bounds, hough_max, make_bins, vote_points, vote_bins, find_voters_points, find_voters_bins
    else:
        from .cpu_impl import projected_bounds, hough_max, make_bins, vote_points, vote_bins, find_voters_points, find_voters_bins

    def find_clusters_points(X, hough, directions, x_min, y_min, dx, dy, reference_time, n=10):
        results = np.full((n, 4), -1)
        results_points = []
        include = np.full(X.shape[0], True)
        for i in range(n):
            idx, val = hough_max(hough)
            print("cluster has value", val, "at", idx)
            voters = find_voters_points(
                hough, X, directions.b, x_min, y_min, dx, dy, reference_time, *idx
            )
            mask = include & voters
            hough = vote_points(
                hough, X[mask], directions.b, x_min, y_min, dx, dy, reference_time, -1
            )
            include &= ~voters # exclude voters
            print(include.sum(), "/", X.shape[0], "points remain")
            results[i, 0] = idx[0]
            results[i, 1] = idx[1]
            results[i, 2] = idx[2]
            results[i, 3] = val
            results_points.append(X[mask])
            
        return results, results_points

    def find_clusters_bins(X, bins, hough, n=10):
        results = np.full((n, 4), -1)
        results_points = []
        include = np.full(bins.shape[0], True)
        for i in range(n):
            idx, val = hough_max(hough)
            print("cluster has value", val, "at", idx)
            voters = find_voters_bins(
                hough, bins, *idx
            )
            mask = include & voters
            hough = vote_bins(
                hough, bins[mask], -1
            )
            include &= ~voters # exclude voters
            print(include.sum(), "/", X.shape[0], "points remain")
            results[i, 0] = idx[0]
            results[i, 1] = idx[1]
            results[i, 2] = idx[2]
            results[i, 3] = val
            results_points.append(X[mask])
            
        return results, results_points

    n = X.shape[0]
    
    x_min, x_max, y_min, y_max = projected_bounds(X, directions.b, reference_time)
    
    num_dir = directions.b.shape[0]
    _dx = dx.to(u.deg).value
    _dy = _dx
    num_x = int((x_max - x_min) / _dx  + 1)
    num_y = int((y_max - y_min) / _dy  + 1)

    log.info("creating hough space with shape (%d, %d, %d)", num_dir, num_x, num_y)
    hough = np.zeros((num_dir, num_x, num_y), dtype=np.uint32)    
    
    if precompute:
        bins = make_bins(X, directions.b, x_min, y_min, _dx, _dy, reference_time)
        hough = vote_bins(hough, bins, 1)
        results, results_points = find_clusters_bins(X, bins, hough, n=num_results)
    else:
        hough = vote_points(hough, X, directions.b, x_min, y_min, _dx, _dy, reference_time, 1)
        results, results_points = find_clusters_points(X, hough, directions, x_min, y_min, _dx, _dy, reference_time, n=num_results)
    
    return results, results_points

def main():
    import argparse
    parser = argparse.ArgumentParser(prog="find-asteroids", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--catalog", required=True, type=Path, help="The detection catalog to search. An astropy-readable table containing at least 'ra', 'dec', and 'time' columns (with units).")
    parser.add_argument("--psfs", required=False, default=None, type=Path, help="An astropy-readable table containing a 'psf' column (with units) that specifies the PSF-widths of the images from which the detection catalog is derived. If not provided, a value of 1 arcsec is assumed.")
    parser.add_argument("--velocity", required=True, nargs=2, type=float, help="The velocity range over which to search, in units of deg/day.")
    parser.add_argument("--angle", required=True, nargs=2, type=float, help="The on-sky angles over which to search, in units of deg.")
    parser.add_argument("--dx", required=True, type=float, help="Search bin-width, in units of the PSF-width.")
    parser.add_argument("--num-results", required=True, type=int, help="Number of results to produce.")
    parser.add_argument("--results-dir", type=Path, required=True, help="The directory into which to write results.")
    parser.add_argument("--precompute", action='store_true', help="Precompute projected positions of detections for all trial velocities (uses more memory, but may be faster).")
    parser.add_argument("--gpu", action='store_true', help="Run the core-search components of the algorithm on GPU.")
    parser.add_argument("--gpu-kernels", action='store_true', help="Run the entirety of the search algorithm on the GPU.")
    parser.add_argument("--device", type=int, required=False, default=-1, help="The GPU device number to use.")
    parser.add_argument("--output-format", type=str, default='ecsv', help="The astropy.table supported format for writing results.")

    args = parser.parse_args()

    if args.gpu and args.device > -1:
        cuda.select_device(args.device)
    
    if args.psfs:
        psfs = astropy.table.Table.read(args.psfs)['psf']
        log.info("seeing [min, median, max]: [%f, %f, %f]", np.min(psfs), np.median(psfs), np.max(psfs))
        psf_scaling = np.median(psfs)
    else:
        psf_scaling = 1 * u.arcsec

    dx = args.dx * psf_scaling * psfs.unit
    log.info(f"using dx = {dx}")
    catalog = astropy.table.Table.read(args.catalog)

    X = np.array([catalog['ra'], catalog['dec'], catalog['time']]).T
    reference_epoch = X[:, 2].min() * u.day
    dt = (X[:, 2].max() - X[:, 2].min()) * u.day
    vmin = args.velocity[0] * u.deg/u.day
    vmax = args.velocity[1] * u.deg/u.day
    phimin = args.angle[0] * u.deg
    phimax = args.angle[1] * u.deg

    directions = SearchDirections([vmin, vmax], [phimin, phimax], dx, dt)
    log.info("searching %d directions", len(directions.b))
    if args.gpu_kernels:
        results, results_points = search_gpu(X, directions, dx, reference_epoch.value, num_results=args.num_results)
    else:
        results, results_points = search(X, directions, dx, reference_epoch.value, num_results=args.num_results, precompute=args.precompute, gpu=args.gpu)

    args.results_dir.mkdir(parents=True, exist_ok=False)
    for i, (result, points) in enumerate(zip(results, results_points)):
        # refine
        try:
            mcdr = refine(points)
            gathered = gather(mcdr, X[:, 0], X[:, 1], X[:, 2], 1/3600)
        except Exception as e:
            log.error(str(e))
            continue
        d = args.results_dir / str(i)
        d.mkdir(parents=True, exist_ok=True)
        astropy.table.Table(
            [
                {
                    "x": result[0],
                    "y": result[1],
                    "direction": result[2],
                    "n": result[3],
                }
            ]
        ).write(d / f"result.{args.output_format}")

        reference_sky_pos = mcdr.predict(np.atleast_2d([reference_epoch.value]))
        astropy.table.Table(
            [
                {
                    "vra": mcdr.beta[0, 0] * u.deg/u.day,
                    "vdec": mcdr.beta[0, 1] * u.deg/u.day,
                    "ra_0": mcdr.alpha[0] * u.deg,
                    "dec_0": mcdr.alpha[1] * u.deg,
                    "ra_ref": reference_sky_pos[0][0],
                    "dec_ref": reference_sky_pos[0][1],
                    "tref": reference_epoch,
                    "tmin": points[:, 2].min() * u.day,
                    "tmax": points[:, 2].max() * u.day,
                    "sigma_vra": mcdr.sigma_e[0, 0]**0.5 * u.deg/u.day,
                    "sigma_vdec": mcdr.sigma_e[1, 1]**0.5 * u.deg/u.day,
                    "sigma_vravdec": mcdr.sigma_e[0, 1] * (u.deg/u.day)**2,
                    "sigma_vdecvra": mcdr.sigma_e[1, 0] * (u.deg/u.day)**2,
                    "sigma_t": mcdr.sigma_xx[0, 0] * u.day,
                }
            ]
        ).write(d / f"tracklet.{args.output_format}")

        t = astropy.table.Table(
            [
                {
                    "ra": p[0] * u.deg,
                    "dec": p[1] * u.deg,
                    "time": p[2] * u.day,
                }
                for p in points
            ]
        )
        t.sort("time")
        t.write(d / f"points.{args.output_format}")

        t = catalog[gathered]
        t.sort("time")
        t.write(d / f"gathered.{args.output_format}")


if __name__ == "__main__":
    main()



import numpy as np
import astropy.units as u

class SearchDirections(object):
    def __init__(self, velocity_range, angle_range, dx, dt, b=None):
        """
        specify two ranges and units
        """
        if b is not None:
            self._b = b
            _v = ((b**2).sum(axis=1)**0.5)
            _phi = np.arctan2(b[:, 1], b[:, 0]) # this provides values in [-pi, pi]
            self.velocity_range = [_v.min(), _v.max()] * b.unit
            self.angle_range = [_phi.min(), _phi.max()] * u.rad
        else:
            self.velocity_range = velocity_range
            # self.angle_range = angle_range
            # fit into [-pi, pi]
            self.angle_range = [
                (angle_range[0] + np.pi*u.rad) % (2 * np.pi * u.rad) - np.pi * u.rad,
                (angle_range[1] + np.pi*u.rad) % (2 * np.pi * u.rad) - np.pi * u.rad
            ]
        
        self.dx = dx
        self.dt = dt

    @property
    def v_min(self):
        return min(self.velocity_range)
    
    @property
    def v_max(self):
        return max(self.velocity_range)
        
    @property
    def phi_min(self):
        return min(self.angle_range)
    
    @property
    def phi_max(self):
        return max(self.angle_range)

        
    @property
    def b(self):
        if not hasattr(self, "_b"):
            _min = (-self.v_max * self.dt/self.dx).to(u.dimensionless_unscaled)
            _max = (self.v_max * self.dt/self.dx).to(u.dimensionless_unscaled)
            _v_x = np.arange(_min, _max + 1, 1)
            _v_y = np.arange(_min, _max + 1, 1)
            v_x_m, v_y_m = np.meshgrid(_v_x, _v_y)
            b = (np.vstack([v_x_m.flatten(), v_y_m.flatten()]).T * (self.dx/self.dt)).to(self.v_max.unit)
            _v = ((b**2).sum(axis=1)**0.5)
            _phi = np.arctan2(b[:, 1], b[:, 0]) # this provides values in [-pi, pi]
            
            v_mask = (_v >= self.v_min) & (_v <= self.v_max)
            # how do I slice a region of [-pi, pi] that resepcts [phi_min, phi_max]?
            if (self.angle_range[1] < self.angle_range[0]) and (self.angle_range[0] < 180*u.deg):
                phi_mask = (_phi >= self.angle_range[0]) | (_phi <= self.angle_range[1]) 
            else:
                phi_mask = (_phi >= self.angle_range[0]) & (_phi <= self.angle_range[1])

            b_m = b.reshape(len(_v_x), len(_v_y), 2)
            adjacent_dx_width = (b_m - np.roll(np.roll(b_m, 1, axis=0), 1, axis=1))[1:, 1:] * self.dt/self.dx
            assert(np.allclose(adjacent_dx_width[~np.isnan(adjacent_dx_width)], 1)) # all adjacent velocities are within 1 dx over timespan of dt
            self._b = b[v_mask & phi_mask]
        return self._b




import logging
import numpy as np
from mcd_regression import MCDRegression
import sys
import astropy.table

logging.basicConfig()
log = logging.getLogger(__name__)

def refine(points):
    x, y = points[:, 2][:, None], points[:, :2]
    log.info(f"refining cluster with {len(points)} points")
    if len(points) < 2:
        log.warn("cluster has too few points to fit a line")
        return None
    try:
        try:
            mcdr = MCDRegression()
            mcdr.fit(x, y)
        except AssertionError as e:
            log.warn("regression on points in cluster failed %s", e)
            return None
    except Exception as e:
        log.exception(e)
        return None
    
    regression_error = (np.diag(mcdr.sigma_e)**2).sum()
    log.debug(f"regression error {regression_error}")
    outliers = mcdr.outliers_r
    inliers = ~outliers
    log.info(f"refined cluster has {inliers.sum()} inliers")

    return mcdr

def gather(mcdr : MCDRegression, ra, dec, time, threshold):
    """
    gather all points within some threshold of a line
    """
    y_pred = mcdr.predict(time[:, None])
    residuals = (
        np.array([ra, dec]).T - 
        y_pred
    )
    distance = (residuals**2).sum(axis=1)**0.5
    mask = distance < threshold
    log.info("gathered %s points", mask.sum())
    return mask


from numba import cuda
import numpy as np
from .primitives import project_xyz, digitize_point

@cuda.jit
def _hough_max(result, hough):
    """Find the maximum value in values and store in result[0]"""
    tid = cuda.threadIdx.x
    bid = cuda.blockIdx.x
    bdim = cuda.blockDim.x
    i = (bid * bdim) + tid
    
    if i >= hough.shape[0]:
        return
    
    y = -1
    z = -1
    v = -np.inf
    for j in range(hough.shape[1]):
        for k in range(hough.shape[2]):
            val = hough[i, j, k]
            if val > v:
                v = val
                y = j
                z = k
    
    result[i][0] = y
    result[i][1] = z
    result[i][2] = v

def hough_max(hough):
    result = np.zeros((hough.shape[0], 3), dtype=np.int32)
    _hough_max[512, hough.shape[0] // 512 + 1](result, hough)
    i = result[:, 2].argmax()
    j = result[i, 0]
    k = result[i, 1]
    v = result[i, 2]
    return (i, j, k), v

@cuda.jit
def _projected_bounds(X, directions, reference_time, min_x_arr, max_x_arr, min_y_arr, max_y_arr):
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    n, d = X.shape
    num_dir = directions.shape[0]
    
    if dir_idx >= num_dir:
        return

    direction = directions[dir_idx]
    
    # Initialize local min/max
    min_x = np.inf
    max_x = -np.inf
    min_y = np.inf
    max_y = -np.inf

    # Iterate over all points
    for i in range(n):
        p = X[i]
        x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
        min_x = min(min_x, x_prime)
        max_x = max(max_x, x_prime)
        min_y = min(min_y, y_prime)
        max_y = max(max_y, y_prime)

    # Store results in global memory
    min_x_arr[dir_idx] = min_x
    max_x_arr[dir_idx] = max_x
    min_y_arr[dir_idx] = min_y
    max_y_arr[dir_idx] = max_y
    
def projected_bounds(X: np.ndarray, directions: np.ndarray, reference_time: np.float64):
    n, d = X.shape
    num_dir = directions.shape[0]

    # Allocate device arrays
    d_X = cuda.to_device(X)
    d_directions = cuda.to_device(directions)
    d_min_x_arr = cuda.device_array(num_dir, dtype=np.float64)
    d_max_x_arr = cuda.device_array(num_dir, dtype=np.float64)
    d_min_y_arr = cuda.device_array(num_dir, dtype=np.float64)
    d_max_y_arr = cuda.device_array(num_dir, dtype=np.float64)

    # Define CUDA thread/block config
    threads_per_block = 128
    blocks_per_grid = (num_dir + threads_per_block - 1) // threads_per_block

    # Launch kernel
    print(blocks_per_grid, threads_per_block)
    _projected_bounds[blocks_per_grid, threads_per_block](
        d_X, d_directions, reference_time, d_min_x_arr, d_max_x_arr, d_min_y_arr, d_max_y_arr
    )

    # Copy results back to host
    min_x_arr = d_min_x_arr.copy_to_host()
    max_x_arr = d_max_x_arr.copy_to_host()
    min_y_arr = d_min_y_arr.copy_to_host()
    max_y_arr = d_max_y_arr.copy_to_host()

    # Final reduction on CPU
    min_x = np.min(min_x_arr)
    max_x = np.max(max_x_arr)
    min_y = np.min(min_y_arr)
    max_y = np.max(max_y_arr)

    return min_x, max_x, min_y, max_y

@cuda.jit
def _make_bins(bins, X, directions, x_min, y_min, dx, dy, reference_time):
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    point_idx = cuda.blockIdx.y * cuda.blockDim.y + cuda.threadIdx.y

    num_dir, n = directions.shape[0], X.shape[0]

    if dir_idx >= num_dir or point_idx >= n:
        return  # Out-of-bounds check

    direction = directions[dir_idx]
    p = X[point_idx]
    x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
    x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)
    bins[point_idx, dir_idx, 0] = x_idx
    bins[point_idx, dir_idx, 1] = y_idx


def make_bins(X, directions, x_min, y_min, dx, dy, reference_time):
    n, d = X.shape
    num_dir = directions.shape[0]
    bins = np.full((n, num_dir, 2), -1)

    # Transfer data to GPU
    d_bins = cuda.to_device(bins)
    d_X = cuda.to_device(X)
    d_directions = cuda.to_device(directions)

    # Configure GPU threads and blocks
    threads_per_block = (16, 16)  # Tunable parameters
    blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                       (n + threads_per_block[1] - 1) // threads_per_block[1])
    
    print(blocks_per_grid, threads_per_block)
    _make_bins[blocks_per_grid, threads_per_block](
        d_bins, d_X, d_directions, x_min, y_min, dx, dy, reference_time
    )

    # Copy back the updated Hough space
    return d_bins.copy_to_host()


@cuda.jit
def _vote_points(hough, X, directions, x_min, y_min, dx, dy, reference_time, coef):
    """
    GPU version of vote_points using CUDA parallelism.
    Each thread processes one (dir_idx, point_idx) pair.
    """
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    point_idx = cuda.blockIdx.y * cuda.blockDim.y + cuda.threadIdx.y

    num_dir, n = directions.shape[0], X.shape[0]

    if dir_idx >= num_dir or point_idx >= n:
        return  # Out-of-bounds check

    direction = directions[dir_idx]
    p = X[point_idx]
    x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
    x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)

    if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
        cuda.atomic.add(hough, (dir_idx, x_idx, y_idx), coef)  # Atomic update to avoid race conditions

@cuda.jit
def _vote_points_mask(hough, mask, X, directions, x_min, y_min, dx, dy, reference_time, coef):
    """
    GPU version of vote_points using CUDA parallelism.
    Each thread processes one (dir_idx, point_idx) pair.
    """
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    point_idx = cuda.blockIdx.y * cuda.blockDim.y + cuda.threadIdx.y
    
    include = False
    for i in range(mask.shape[0]):
        if mask[i, point_idx]:
            include = True
    
    if not include:
        return

    num_dir, n = directions.shape[0], X.shape[0]

    if dir_idx >= num_dir or point_idx >= n:
        return  # Out-of-bounds check

    direction = directions[dir_idx]
    p = X[point_idx]
    x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
    x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)

    if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
        cuda.atomic.add(hough, (dir_idx, x_idx, y_idx), coef)  # Atomic update to avoid race conditions


def vote_points(hough: np.ndarray, X: np.ndarray, directions: np.ndarray, x_min, y_min, dx, dy, reference_time, coef: np.float64 = 1):
    """
    Host function to launch the GPU kernel.
    """
    n, d = X.shape
    num_dir = directions.shape[0]

    # Transfer data to GPU
    d_hough = cuda.to_device(hough)
    d_X = cuda.to_device(X)
    d_directions = cuda.to_device(directions)

    # Configure GPU threads and blocks
    threads_per_block = (16, 16)  # Tunable parameters
    blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                       (n + threads_per_block[1] - 1) // threads_per_block[1])

    print(blocks_per_grid, threads_per_block)
    _vote_points[blocks_per_grid, threads_per_block](
        d_hough, d_X, d_directions, x_min, y_min, dx, dy, reference_time, coef
    )

    # Copy back the updated Hough space
    return d_hough.copy_to_host()

@cuda.jit
def _vote_bins(hough, bins, coef):
    """
    GPU version of vote_points using CUDA parallelism.
    Each thread processes one (dir_idx, point_idx) pair.
    """
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    point_idx = cuda.blockIdx.y * cuda.blockDim.y + cuda.threadIdx.y

    n, num_dir, _ = bins.shape

    if dir_idx >= num_dir or point_idx >= n:
        return  # Out-of-bounds check

    x_idx = bins[point_idx, dir_idx, 0]
    y_idx = bins[point_idx, dir_idx, 1]

    if 0 <= x_idx < hough.shape[1] and 0 <= y_idx < hough.shape[2]:
        cuda.atomic.add(hough, (dir_idx, x_idx, y_idx), coef)  # Atomic update to avoid race conditions


def vote_bins(hough: np.ndarray, bins: np.ndarray, coef: np.float64 = 1):
    """
    Host function to launch the GPU kernel.
    """
    n, num_dir, _ = bins.shape

    # Transfer data to GPU
    d_hough = cuda.to_device(hough)
    d_bins = cuda.to_device(bins)

    # Configure GPU threads and blocks
    threads_per_block = (16, 16)  # Tunable parameters
    blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                       (n + threads_per_block[1] - 1) // threads_per_block[1])

    print(blocks_per_grid, threads_per_block)
    _vote_bins[blocks_per_grid, threads_per_block](
        d_hough, d_bins, coef
    )

    # Copy back the updated Hough space
    return d_hough.copy_to_host()

@cuda.jit
def _find_voters_bins(bins, mask, mask_dir, mask_x, mask_y):
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    if dir_idx != mask_dir:
        return
    
    point_idx = cuda.blockIdx.y * cuda.blockDim.y + cuda.threadIdx.y

    n, num_dir, _ = bins.shape

    if dir_idx >= num_dir or point_idx >= n:
        return  # Out-of-bounds check

    x_idx = bins[point_idx, dir_idx, 0]
    y_idx = bins[point_idx, dir_idx, 1]

    if x_idx == mask_x and y_idx == mask_y:
        mask[dir_idx, point_idx] = True

def find_voters_bins(hough: np.ndarray, bins: np.ndarray, mask_dir, mask_x, mask_y):
    n, num_dir, _ = bins.shape

    mask = np.full((num_dir, n), False)

    # Transfer data to GPU
    d_bins = cuda.to_device(bins)
    d_mask = cuda.to_device(mask)

    # Configure GPU threads and blocks
    threads_per_block = (16, 16)  # Tunable parameters
    blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                       (n + threads_per_block[1] - 1) // threads_per_block[1])

    print(blocks_per_grid, threads_per_block)
    _find_voters_bins[blocks_per_grid, threads_per_block](
        d_bins, d_mask, mask_dir, mask_x, mask_y
    )

    return np.logical_or.reduce(d_mask.copy_to_host(), axis=0)

@cuda.jit
def _find_voters_points(X, directions, mask, x_min, y_min, dx, dy, reference_time, mask_dir, mask_x, mask_y):
    dir_idx = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    if dir_idx != mask_dir:
        return
    
    point_idx = cuda.blockIdx.y * cuda.blockDim.y + cuda.threadIdx.y

    n, _ = X.shape
    num_dir = directions.shape[0]

    if dir_idx >= num_dir or point_idx >= n:
        return  # Out-of-bounds check

    direction = directions[dir_idx]
    p = X[point_idx]
    x_prime, y_prime = project_xyz(p[0], p[1], p[2], direction[0], direction[1], reference_time)
    x_idx, y_idx = digitize_point(x_prime, y_prime, x_min, y_min, dx, dy)

    if x_idx == mask_x and y_idx == mask_y:
        mask[dir_idx, point_idx] = True


def find_voters_points(hough, X, directions, x_min, y_min, dx, dy, reference_time, mask_dir, mask_x, mask_y):
    n, d = X.shape
    num_dir = directions.shape[0]
    mask = np.full((num_dir, n), False)
    # Transfer data to GPU
    d_X = cuda.to_device(X)
    d_directions = cuda.to_device(directions)
    d_mask = cuda.to_device(mask)

    # Configure GPU threads and blocks
    threads_per_block = (16, 16)  # Tunable parameters
    blocks_per_grid = ((num_dir + threads_per_block[0] - 1) // threads_per_block[0], 
                       (n + threads_per_block[1] - 1) // threads_per_block[1])

    print(blocks_per_grid, threads_per_block)
    _find_voters_points[blocks_per_grid, threads_per_block](
        d_X, d_directions, d_mask, x_min, y_min, dx, dy, reference_time, mask_dir, mask_x, mask_y
    )

    return np.logical_or.reduce(d_mask.copy_to_host(), axis=0)


